#include "../include/lass.h"

/**
 *
 * 	@file ddss_dsyrk.c
 *
 * 	@brief LASs-DDSs ddss_dsyrk routine.
 *
 * 	LASs-DDSs is a software package provided by:
 * 	Barcelona Supercomputing Center - Centro Nacional de Supercomputacion
 *
 * 	@author Pedro Valero-Lara   pedro.valero@bsc.es
 * 	@author Boro Sofranac       boro.sofranac@bsc.es
 * 	@date 2018-02-13
 * 	@reviewer 
 * 	@modified 
 *
 **/

/**
 *  
 *	@ingroup DDSS
 *   
 *	Performs one of the symmetric rank k operations:
 *
 *		C = ALPHA * A * op( A ) + BETA * C	or
 *		C = ALPHA * op( A ) * A + BETA * C
 *	
 *	where op( X ) is:
 *
 *		op( X ) = X**T
 *
 *	ALPHA and BETA are scalars, C is an N by N symmetric matrix and A is an
 *	N by K matrix in the first case and a K by N matrix in the second case.
 *
 **/

/**
 *	@param[in]
 *	UPLO	enum DDSS_UPLO.
 *	    	UPLO specifies the form in which C is stored:
 *       	- Lower: Lower triangular part of C is stored. The upper traingular
 *       	part is not referenced.
 *      	- Upper: Upper triangular part of C is stored. The lower triangular
 *      	part is not referenced.
 *  
 *  @param[in]
 *	TRANS_A	enum DDSS_TRANS.
 *	        TRANS_A specifies the operation to be performed as follows:  
 *      	- NoTrans:   C = ALPHA * A * A**T + BETA * C
 *       	- Trans:     C = ALPHA * A**T * A + BETA * C
 *       	
 *	@param[in]
 *	N       int.
 *			N specifies the order of matrix C. N must be at least zero.
 *
 *	@param[in]
 *	K       int.
 *          With TRANS_A = NoTrans, K specifies the number of columns of the
 *          matrix A, and with TRANS_A = Trans, K specifies the number of rows 
 *			of the matrix A. K must be at least zero.
 *    
 *	@param[in]
 *  ALPHA   double.
 *   		ALPHA specifies the scalar alpha.
 *   		
 *	@param[in]
 *  A  		double *.
 *          A is a pointer to a matrix of dimension Na ( rows ) by Ka  
 *			( columns ), where Na is N and Ka is K when TRANS_A = NoTrans, 
 *			and Na is K and Ka is N otherwise.
 *
 *  @param[in]
 *  LDA     int.
 *          LDA specifies the number of columns of A ( row-major order ).
 *          When TRANS_A = NoTrans then LDA must be at least  max( 1, K ),
 *          otherwise LDA must be at least max( 1, N ).
 *          
 *	@param[in]
 *  BETA    double.
 *   		BETA specifies the scalar beta.
 *   
 *	@param[in, out]
 *  C  		double *.
 * 		    C is a pointer to a matrix of dimension N by N. 
 *			When UPLO = Uppper the strictly lower triangular part of C is not
 *			referenced. On exit, the upper triangular part of C is overwritten
 *			by the upper triangular part of the updated solution matrix C.
 *			When UPLO = Lower the strictly upper triangular part of C is not
 *			referenced. On exit, the lower triangular part of C is overwritten
 *			by the lower triangular part of the updated solution matrix C.
 *   
 *	@param[in]
 *  LDC     int.
 * 		    LDC specifies the number of columns of C ( row-major order ).
 *          LDC must be at least max( 1, N ).
 *
 **/

/**
 *
 *	@sa kdsyrk
 * 
 **/

int ddss_dsyrk( enum DDSS_UPLO UPLO, enum DDSS_TRANS TRANS_A, 
		int N, int K,
        const double ALPHA, double *A, int LDA,
    	const double BETA,  double *C, int LDC )
{

	// Local variables
	int nA;
	
	// Argument checking
	if ( ( UPLO != Upper ) && ( UPLO != Lower ) ) 
	{
		fprintf( stderr, "Illegal value of UPLO, in ddss_dsyrk code\n" );
		return NoSuccess;
	}
	
	if ( ( TRANS_A != NoTrans ) && ( TRANS_A != Trans ) )
	{
		fprintf( stderr, "Illegal value of TRANS_A, in ddss_dsyrk code\n" );
		return NoSuccess;
	}
	
	if ( N < 0 )
	{
		fprintf( stderr, "Illegal value of N, in ddss_dsyrk code\n" );
		return NoSuccess;
	}
		
	if ( K < 0 )
	{
		fprintf( stderr, "Illegal value of K, in ddss_dsyrk code\n" );
		return NoSuccess;
	}
	
	if ( TRANS_A == NoTrans ) 
	{
		nA = K;
	} 
	else 
	{
		nA = N;
	}

	if ( LDA < MAX( 1, nA ) ) 
	{
		fprintf( stderr, "Illegal value of LDA, in ddss_dsyrk code\n" );
		return NoSuccess;
	}

	if ( LDC < MAX( 1, N ) ) 
	{
		fprintf( stderr, "Illegal value of LDC, in ddss_dsyrk code\n" );
		return NoSuccess;
	}
	
	// Quick return
	 if ( N == 0 || ( ( ALPHA == 0.0 || K == 0 ) && BETA == 1.0 ) )
	 {
		 return Success;
	 }
	
	return kdsyrk( UPLO, TRANS_A,
				   N, K,
				   ALPHA, A, LDA,
				   BETA,  C, LDC );
	
}
